@@ -76,19 +76,19 @@ check_cancelled(void *arg)
 struct gotadmin_cmd {
 	const char	*cmd_name;
 	const struct got_error *(*cmd_main)(int, char *[]);
-	void		(*cmd_usage)(void);
+	void		(*cmd_usage)(int);
 	const char	*cmd_alias;
 };
 
 __dead static void	usage(int, int);
-__dead static void	usage_init(void);
-__dead static void	usage_info(void);
-__dead static void	usage_pack(void);
-__dead static void	usage_indexpack(void);
-__dead static void	usage_listpack(void);
-__dead static void	usage_cleanup(void);
-__dead static void	usage_dump(void);
-__dead static void	usage_load(void);
+__dead static void	usage_init(int);
+__dead static void	usage_info(int);
+__dead static void	usage_pack(int);
+__dead static void	usage_indexpack(int);
+__dead static void	usage_listpack(int);
+__dead static void	usage_cleanup(int);
+__dead static void	usage_dump(int);
+__dead static void	usage_load(int);
 
 static const struct got_error*		cmd_init(int, char *[]);
 static const struct got_error*		cmd_info(int, char *[]);
@@ -177,7 +177,7 @@ main(int argc, char *argv[])
 			continue;
 
 		if (hflag)
-			cmd->cmd_usage();
+			cmd->cmd_usage(0);
 
 		error = cmd->cmd_main(argc, argv);
 		if (error && error->code != GOT_ERR_CANCELLED &&
@@ -236,11 +236,12 @@ apply_unveil(const char *repo_path, int repo_read_only)
 }
 
 __dead static void
-usage_info(void)
+usage_info(int status)
 {
-	fprintf(stderr, "usage: %s info [-r repository-path]\n",
+	FILE *fp = (status == 0) ? stdout : stderr;
+	fprintf(fp, "usage: %s info [-r repository-path]\n",
 	    getprogname());
-	exit(1);
+	exit(status);
 }
 
 static const struct got_error *
@@ -277,12 +278,13 @@ done:
 }
 
 __dead static void
-usage_init(void)
+usage_init(int status)
 {
-	fprintf(stderr, "usage: %s init [-A hashing-algorithm] [-b branch]"
+	FILE *fp = (status == 0) ? stdout : stderr;
+	fprintf(fp, "usage: %s init [-A hashing-algorithm] [-b branch]"
 	    " repository-path\n",
 	    getprogname());
-	exit(1);
+	exit(status);
 }
 
 static const struct got_error *
@@ -314,7 +316,7 @@ cmd_init(int argc, char *argv[])
 			head_name = optarg;
 			break;
 		default:
-			usage_init();
+			usage_init(1);
 			/* NOTREACHED */
 		}
 	}
@@ -323,7 +325,7 @@ cmd_init(int argc, char *argv[])
 	argv += optind;
 
 	if (argc != 1)
-		usage_init();
+		usage_init(1);
 
 	repo_path = strdup(argv[0]);
 	if (repo_path == NULL)
@@ -374,7 +376,7 @@ cmd_info(int argc, char *argv[])
 			got_path_strip_trailing_slashes(repo_path);
 			break;
 		default:
-			usage_info();
+			usage_info(1);
 			/* NOTREACHED */
 		}
 	}
@@ -472,11 +474,12 @@ done:
 }
 
 __dead static void
-usage_pack(void)
+usage_pack(int status)
 {
-	fprintf(stderr, "usage: %s pack [-aDq] [-r repository-path] "
+	FILE *fp = (status == 0) ? stdout : stderr;
+	fprintf(fp, "usage: %s pack [-aDq] [-r repository-path] "
 	    "[-x reference] [reference ...]\n", getprogname());
-	exit(1);
+	exit(status);
 }
 
 struct got_pack_progress_arg {
@@ -518,7 +521,7 @@ print_load_info(FILE *out, int print_colored, int print_found, int print_trees,
 static const struct got_error *
 pack_progress(void *arg, int ncolored, int nfound, int ntrees,
     off_t packfile_size, int ncommits, int nobj_total, int nobj_deltify,
-    int nobj_written)
+    int nobj_written, int pack_done)
 {
 	struct got_pack_progress_arg *a = arg;
 	char scaled_size[FMT_SCALED_STRSIZE];
@@ -619,17 +622,18 @@ pack_progress(void *arg, int ncolored, int nfound, int ntrees,
 	if (print_written)
 		fprintf(a->out, "; writing pack: %*s %d%%",
 		    FMT_SCALED_STRSIZE - 2, scaled_size, p_written);
-	if (print_searching || print_total || print_deltify ||
-	    print_written) {
+	if (print_searching || print_total || print_deltify || print_written) {
 		a->printed_something = 1;
 		fflush(a->out);
 	}
+	if (pack_done)
+		fprintf(a->out, "\n");
 	return NULL;
 }
 
 static const struct got_error *
 pack_index_progress(void *arg, off_t packfile_size, int nobj_total,
-    int nobj_indexed, int nobj_loose, int nobj_resolved)
+    int nobj_indexed, int nobj_loose, int nobj_resolved, int indexing_done)
 {
 	struct got_pack_progress_arg *a = arg;
 	char scaled_size[FMT_SCALED_STRSIZE];
@@ -676,7 +680,9 @@ pack_index_progress(void *arg, off_t packfile_size, int nobj_total,
 		printf("; indexing %d%%", p_indexed);
 	if (print_resolved)
 		printf("; resolving deltas %d%%", p_resolved);
-	if (print_size || print_indexed || print_resolved)
+	if (indexing_done)
+		printf("\n");
+	if (print_size || print_indexed || print_resolved || indexing_done)
 		fflush(stdout);
 
 	return NULL;
@@ -717,7 +723,7 @@ cmd_pack(int argc, char *argv[])
 	struct got_repository *repo = NULL;
 	int ch, i, loose_obj_only = 1, force_refdelta = 0, verbosity = 0;
 	struct got_object_id *pack_hash = NULL;
-	char *id_str = NULL;
+	char *id_str = NULL, *idxpath = NULL;
 	struct got_pack_progress_arg ppa;
 	FILE *packfile = NULL;
 	struct got_pathlist_head exclude_args;
@@ -763,7 +769,7 @@ cmd_pack(int argc, char *argv[])
 				return error;
 			break;
 		default:
-			usage_pack();
+			usage_pack(1);
 			/* NOTREACHED */
 		}
 	}
@@ -842,7 +848,7 @@ cmd_pack(int argc, char *argv[])
 	if (verbosity >= 0)
 		printf("\nWrote %s.pack\n", id_str);
 
-	error = got_repo_index_pack(packfile, pack_hash, repo,
+	error = got_repo_index_pack(&idxpath, packfile, pack_hash, repo,
 	    pack_index_progress, &ppa, check_cancelled, NULL);
 	if (error)
 		goto done;
@@ -863,15 +869,17 @@ done:
 	free(id_str);
 	free(pack_hash);
 	free(repo_path);
+	free(idxpath);
 	return error;
 }
 
 __dead static void
-usage_indexpack(void)
+usage_indexpack(int status)
 {
-	fprintf(stderr, "usage: %s indexpack packfile-path\n",
+	FILE *fp = (status == 0) ? stdout : stderr;
+	fprintf(fp, "usage: %s indexpack packfile-path\n",
 	    getprogname());
-	exit(1);
+	exit(status);
 }
 
 static const struct got_error *
@@ -881,7 +889,7 @@ cmd_indexpack(int argc, char *argv[])
 	struct got_repository *repo = NULL;
 	int ch;
 	struct got_object_id *pack_hash = NULL;
-	char *packfile_path = NULL;
+	char *packfile_path = NULL, *idxpath = NULL;
 	char *id_str = NULL;
 	struct got_pack_progress_arg ppa;
 	FILE *packfile = NULL;
@@ -890,7 +898,7 @@ cmd_indexpack(int argc, char *argv[])
 	while ((ch = getopt(argc, argv, "")) != -1) {
 		switch (ch) {
 		default:
-			usage_indexpack();
+			usage_indexpack(1);
 			/* NOTREACHED */
 		}
 	}
@@ -899,7 +907,7 @@ cmd_indexpack(int argc, char *argv[])
 	argv += optind;
 
 	if (argc != 1)
-		usage_indexpack();
+		usage_indexpack(1);
 
 	packfile_path = realpath(argv[0], NULL);
 	if (packfile_path == NULL)
@@ -937,7 +945,7 @@ cmd_indexpack(int argc, char *argv[])
 	if (error)
 		goto done;
 
-	error = got_repo_index_pack(packfile, pack_hash, repo,
+	error = got_repo_index_pack(&idxpath, packfile, pack_hash, repo,
 	    pack_index_progress, &ppa, check_cancelled, NULL);
 	if (error)
 		goto done;
@@ -953,15 +961,17 @@ done:
 	}
 	free(id_str);
 	free(pack_hash);
+	free(idxpath);
 	return error;
 }
 
 __dead static void
-usage_listpack(void)
+usage_listpack(int status)
 {
-	fprintf(stderr, "usage: %s listpack [-hs] packfile-path\n",
+	FILE *fp = (status == 0) ? stdout : stderr;
+	fprintf(fp, "usage: %s listpack [-hs] packfile-path\n",
 	    getprogname());
-	exit(1);
+	exit(status);
 }
 
 struct gotadmin_list_pack_cb_args {
@@ -1081,7 +1091,7 @@ cmd_listpack(int argc, char *argv[])
 			show_stats = 1;
 			break;
 		default:
-			usage_listpack();
+			usage_listpack(1);
 			/* NOTREACHED */
 		}
 	}
@@ -1090,7 +1100,7 @@ cmd_listpack(int argc, char *argv[])
 	argv += optind;
 
 	if (argc != 1)
-		usage_listpack();
+		usage_listpack(1);
 	packfile_path = realpath(argv[0], NULL);
 	if (packfile_path == NULL)
 		return got_error_from_errno2("realpath", argv[0]);
@@ -1149,11 +1159,12 @@ done:
 }
 
 __dead static void
-usage_cleanup(void)
+usage_cleanup(int status)
 {
-	fprintf(stderr, "usage: %s cleanup [-anpq] [-r repository-path]\n",
+	FILE *fp = (status == 0) ? stdout : stderr;
+	fprintf(fp, "usage: %s cleanup [-anpq] [-r repository-path]\n",
 	    getprogname());
-	exit(1);
+	exit(status);
 }
 
 struct got_cleanup_progress_arg {
@@ -1264,6 +1275,7 @@ cmd_cleanup(int argc, char *argv[])
 	int ch, dry_run = 0, verbosity = 0;
 	int ncommits = 0, nloose = 0, npacked = 0;
 	int remove_lonely_packidx = 0, ignore_mtime = 0;
+	struct got_pack_progress_arg ppa;
 	struct got_cleanup_progress_arg cpa;
 	struct got_lonely_packidx_progress_arg lpa;
 	off_t loose_before, loose_after;
@@ -1277,7 +1289,7 @@ cmd_cleanup(int argc, char *argv[])
 	int *pack_fds = NULL;
 
 #ifndef PROFILE
-	if (pledge("stdio rpath wpath cpath flock proc exec sendfd unveil",
+	if (pledge("stdio rpath wpath cpath fattr flock proc exec sendfd unveil",
 	    NULL) == -1)
 		err(1, "pledge");
 #endif
@@ -1304,7 +1316,7 @@ cmd_cleanup(int argc, char *argv[])
 			got_path_strip_trailing_slashes(repo_path);
 			break;
 		default:
-			usage_cleanup();
+			usage_cleanup(1);
 			/* NOTREACHED */
 		}
 	}
@@ -1351,11 +1363,16 @@ cmd_cleanup(int argc, char *argv[])
 	cpa.dry_run = dry_run;
 	cpa.verbosity = verbosity;
 
+	memset(&ppa, 0, sizeof(ppa));
+	ppa.out = stdout;
+	ppa.verbosity = verbosity;
+
 	error = got_repo_cleanup(repo, &loose_before, &loose_after,
 	    &pack_before, &pack_after, &ncommits, &nloose, &npacked,
 	    dry_run, ignore_mtime, cleanup_progress, &cpa,
+	    pack_progress, &ppa, pack_index_progress, &ppa,
 	    check_cancelled, NULL);
-	if (cpa.printed_something)
+	if (ppa.printed_something || cpa.printed_something)
 		printf("\n");
 	if (error)
 		goto done;
@@ -1410,11 +1427,12 @@ done:
 }
 
 __dead static void
-usage_dump(void)
+usage_dump(int status)
 {
-	fprintf(stderr, "usage: %s dump [-q] [-r repository-path] "
+	FILE *fp = (status == 0) ? stdout : stderr;
+	fprintf(fp, "usage: %s dump [-q] [-r repository-path] "
 	    "[-x reference] [reference]...\n", getprogname());
-	exit(1);
+	exit(status);
 }
 
 static const struct got_error *
@@ -1463,7 +1481,7 @@ cmd_dump(int argc, char *argv[])
 				return error;
 			break;
 		default:
-			usage_dump();
+			usage_dump(1);
 			/* NOTREACHED */
 		}
 	}
@@ -1547,12 +1565,13 @@ cmd_dump(int argc, char *argv[])
 }
 
 __dead static void
-usage_load(void)
+usage_load(int status)
 {
-	fprintf(stderr, "usage: %s load [-nq] [-l bundle-file] "
+	FILE *fp = (status == 0) ? stdout : stderr;
+	fprintf(fp, "usage: %s load [-nq] [-l bundle-file] "
 	    "[-r repository-path] [reference ...]\n",
 	    getprogname());
-	exit(1);
+	exit(status);
 }
 
 static const struct got_error *
@@ -1560,7 +1579,7 @@ load_progress(void *arg, off_t packfile_size, int nobj_total,
     int nobj_indexed, int nobj_loose, int nobj_resolved)
 {
 	return pack_index_progress(arg, packfile_size, nobj_total,
-	    nobj_indexed, nobj_loose, nobj_resolved);
+	    nobj_indexed, nobj_loose, nobj_resolved, 0);
 }
 
 static int
@@ -1715,7 +1734,7 @@ cmd_load(int argc, char *argv[])
 			got_path_strip_trailing_slashes(repo_path);
 			break;
 		default:
-			usage_load();
+			usage_load(1);
 			/* NOTREACHED */
 		}
 	}
