@@ -34,7 +34,9 @@
 #include "got_error.h"
 #include "got_reference.h"
 #include "got_object.h"
+#include "got_path.h"
 
+#include "media.h"
 #include "gotwebd.h"
 #include "log.h"
 #include "tmpl.h"
@@ -65,6 +67,7 @@ auth_shutdown(void)
 
 		config_free_access_rules(&srv->access_rules);
 		config_free_repos(&srv->repos);
+		config_free_websites(&srv->websites);
 		free(srv);
 	}
 
@@ -96,8 +99,10 @@ auth_sighdlr(int sig, short event, void *arg)
 		break;
 	case SIGCHLD:
 		break;
-	case SIGINT:
 	case SIGTERM:
+		/* continue until the parent exits */
+		break;
+	case SIGINT:
 		auth_shutdown();
 		break;
 	default:
@@ -235,7 +240,9 @@ render_error(struct request *c, const struct got_error *error)
 
 	c->t->error = error;
 
-	if (error->code == GOT_ERR_LOGIN_FAILED)
+	if (error->code == GOT_ERR_NOT_FOUND)
+		status = 404;
+	else if (error->code == GOT_ERR_LOGIN_FAILED)
 		status = 401;
 	else
 		status = 400;
@@ -430,7 +437,10 @@ login_error_hint(struct request *c)
 		return got_error(GOT_ERR_LOGIN_FAILED);
 
 	ret = snprintf(msg, sizeof(msg),
-	    "Log in by running: ssh %s@%s \"weblogin %s\"",
+	    "Log in by running: ssh %s%s%s%s@%s \"weblogin %s\"",
+	    srv->login_hint_port[0] ? " -p " : "",
+	    srv->login_hint_port[0] ? srv->login_hint_port : "",
+	    srv->login_hint_port[0] ? " " : "",
 	    srv->login_hint_user, srv->name, srv->name);
 	if (ret == -1 || (size_t)ret >= sizeof(msg))
 		return got_error(GOT_ERR_LOGIN_FAILED);
@@ -445,10 +455,14 @@ process_request(struct request *c)
 	struct gotwebd *env = gotwebd_env;
 	uid_t uid;
 	struct server *srv;
+	struct website *site;
 	struct gotwebd_repo *repo = NULL;
 	enum gotwebd_auth_config auth_config;
+	struct gotwebd_access_rule_list *access_rules = NULL;
 	char *hostname = NULL;
 	const char *identifier = NULL;
+	char *request_path = NULL;
+	int is_repository_request = 0;
 
 	srv = gotweb_get_server(c->fcgi_params.server_name);
 	if (srv == NULL) {
@@ -457,11 +471,38 @@ process_request(struct request *c)
 		goto done;
 	}
 
+	error = gotweb_route_request(&is_repository_request, &site,
+	    &request_path, c);
+	if (error)
+		goto done;
+
+	/*
+	 * Static gotwebd assets (images, CSS, ...) are not protected
+	 * by authentication.
+	 */
+	if (is_repository_request && !got_path_is_root_dir(request_path)) {
+		forward_request(c);
+		free(request_path);
+		return;
+	}
+
+	free(request_path);
+	request_path = NULL;
+
 	auth_config = srv->auth_config;
-	if (c->fcgi_params.qs.path[0] != '\0') {
+
+	if (c->fcgi_params.qs.path[0] != '\0')
 		repo = gotweb_get_repository(srv, c->fcgi_params.qs.path);
-		if (repo)
-			auth_config = repo->auth_config;
+
+	if (repo) {
+		auth_config = repo->auth_config;
+		access_rules = &repo->access_rules;
+	} else if (site) {
+		auth_config = site->auth_config;
+		access_rules = &site->access_rules;
+
+		/* Ignore the querystring while serving web sites. */
+		fcgi_init_querystring(&c->fcgi_params.qs);
 	}
 
 	switch (auth_config) {
@@ -498,8 +539,8 @@ process_request(struct request *c)
 		goto done;
 	}
 
-	if (repo) {
-		switch (auth_check(&identifier, uid, &repo->access_rules)) {
+	if (access_rules) {
+		switch (auth_check(&identifier, uid, access_rules)) {
 		case GOTWEBD_ACCESS_DENIED:
 			error = got_error_msg(GOT_ERR_LOGIN_FAILED,
 			    "permission denied");
@@ -596,6 +637,7 @@ permitted:
 	}
 done:
 	free(hostname);
+	free(request_path);
 	if (error)
 		render_error(c, error);
 	else
@@ -884,6 +926,8 @@ auth_dispatch_main(int fd, short event, void *arg)
 	struct gotwebd		*env = gotwebd_env;
 	struct server		*srv;
 	struct gotwebd_repo	*repo;
+	struct got_pathlist_entry *pe;
+	struct website		*site;
 	ssize_t			 n;
 	int			 shut = 0;
 
@@ -915,9 +959,18 @@ auth_dispatch_main(int fd, short event, void *arg)
 			} else {
 				srv = TAILQ_LAST(&env->servers, serverlist);
 				if (TAILQ_EMPTY(&srv->repos)) {
-					/* per-server access rule */
-					config_get_access_rule(
-					    &srv->access_rules, &imsg);
+					if (RB_EMPTY(&srv->websites)) {
+						/* per-server access rule */
+						config_get_access_rule(
+						    &srv->access_rules, &imsg);
+					} else {
+						pe = RB_MAX(got_pathlist_head,
+						    &srv->websites);
+						site = pe->data;
+						/* per-website access rule */
+						config_get_access_rule(
+						    &site->access_rules, &imsg);
+					}
 				} else {
 					/* per-repository access rule */
 					repo = TAILQ_LAST(&srv->repos,
@@ -936,6 +989,14 @@ auth_dispatch_main(int fd, short event, void *arg)
 			srv = TAILQ_LAST(&env->servers, serverlist);
 			config_get_repository(&srv->repos, &imsg);
 			break;
+		case GOTWEBD_IMSG_CFG_WEBSITE:
+			if (TAILQ_EMPTY(&env->servers)) {
+				fatalx("%s: unexpected CFG_WEBSITE msg",
+				    __func__);
+			}
+			srv = TAILQ_LAST(&env->servers, serverlist);
+			config_get_website(&srv->websites, &imsg);
+			break;
 		case GOTWEBD_IMSG_CTL_PIPE:
 			if (env->iev_sockets == NULL)
 				recv_sockets_pipe(env, &imsg);
