@@ -40,9 +40,11 @@
 #include "got_error.h"
 #include "got_reference.h"
 #include "got_object.h"
+#include "got_path.h"
 
 #include "got_lib_poll.h"
 
+#include "media.h"
 #include "gotwebd.h"
 #include "log.h"
 #include "tmpl.h"
@@ -96,8 +98,10 @@ fcgi_sighdlr(int sig, short event, void *arg)
 		break;
 	case SIGCHLD:
 		break;
-	case SIGINT:
 	case SIGTERM:
+		/* continue until the parent exits */
+		break;
+	case SIGINT:
 		fcgi_shutdown();
 		break;
 	default:
@@ -179,13 +183,14 @@ fcgi_parse_record(struct gotwebd_fcgi_record *rec)
 }
 
 static const struct querystring_keys querystring_keys[] = {
-	{ "action",		ACTION },
-	{ "commit",		COMMIT },
-	{ "file",		RFILE },
-	{ "folder",		FOLDER },
-	{ "headref",		HEADREF },
-	{ "index_page",		INDEX_PAGE },
-	{ "path",		PATH },
+	{ "action",		GOTWEBD_QS_ACTION },
+	{ "commit",		GOTWEBD_QS_COMMIT },
+	{ "file",		GOTWEBD_QS_RFILE },
+	{ "folder",		GOTWEBD_QS_FOLDER },
+	{ "headref",		GOTWEBD_QS_HEADREF },
+	{ "index_page",	GOTWEBD_QS_INDEX_PAGE },
+	{ "path",		GOTWEBD_QS_PATH },
+	{ "login",		GOTWEBD_QS_LOGIN },
 };
 
 static const struct action_keys action_keys[] = {
@@ -293,7 +298,7 @@ assign_querystring(struct querystring *qs, char *key, char *value)
 			continue;
 
 		switch (querystring_keys[el_cnt].element) {
-		case ACTION:
+		case GOTWEBD_QS_ACTION:
 			for (a_cnt = 0; a_cnt < nitems(action_keys); a_cnt++) {
 				if (strcmp(value, action_keys[a_cnt].name) != 0)
 					continue;
@@ -303,7 +308,7 @@ assign_querystring(struct querystring *qs, char *key, char *value)
 			if (a_cnt == nitems(action_keys))
 				qs->action = ERR;
 			break;
-		case COMMIT:
+		case GOTWEBD_QS_COMMIT:
 			if (strlcpy(qs->commit, value, sizeof(qs->commit)) >=
 			    sizeof(qs->commit)) {
 				error = got_error_msg(GOT_ERR_NO_SPACE,
@@ -311,7 +316,7 @@ assign_querystring(struct querystring *qs, char *key, char *value)
 				goto done;
 			}
 			break;
-		case RFILE:
+		case GOTWEBD_QS_RFILE:
 			error = validate_path(value);
 			if (error)
 				goto done;
@@ -322,7 +327,7 @@ assign_querystring(struct querystring *qs, char *key, char *value)
 				goto done;
 			}
 			break;
-		case FOLDER:
+		case GOTWEBD_QS_FOLDER:
 			error = validate_path(value);
 			if (error)
 				goto done;
@@ -334,7 +339,7 @@ assign_querystring(struct querystring *qs, char *key, char *value)
 				goto done;
 			}
 			break;
-		case HEADREF:
+		case GOTWEBD_QS_HEADREF:
 			error = validate_path(value);
 			if (error)
 				goto done;
@@ -345,7 +350,7 @@ assign_querystring(struct querystring *qs, char *key, char *value)
 				goto done;
 			}
 			break;
-		case INDEX_PAGE:
+		case GOTWEBD_QS_INDEX_PAGE:
 			if (*value == '\0')
 				break;
 			qs->index_page = strtonum(value, 0, INT_MAX, &errstr);
@@ -355,7 +360,7 @@ assign_querystring(struct querystring *qs, char *key, char *value)
 				goto done;
 			}
 			break;
-		case PATH:
+		case GOTWEBD_QS_PATH:
 			error = validate_path(value);
 			if (error)
 				goto done;
@@ -366,7 +371,15 @@ assign_querystring(struct querystring *qs, char *key, char *value)
 				goto done;
 			}
 			break;
-		}
+		case GOTWEBD_QS_LOGIN:
+			if (strlcpy(qs->login, value, sizeof(qs->login)) >=
+			    sizeof(qs->login)) {
+				error = got_error_msg(GOT_ERR_NO_SPACE,
+				    "login token too long");
+				goto done;
+			}
+			break;
+ 		}
 
 		/* entry found */
 		break;
@@ -425,6 +438,47 @@ err:
 	return error;
 }
 
+static void
+parse_cookie_hdr(struct gotwebd_fcgi_params *params, char *hdr, size_t len)
+{
+	size_t	 l;
+	char	*end;
+
+	memset(params->auth_cookie, 0, sizeof(params->auth_cookie));
+
+	while (len > 0) {
+		if (hdr[0] == ' ' || hdr[0] == '\t') {
+			hdr++;
+			len--;
+			continue;
+		}
+
+		/* looking at the start of name=val */
+
+		if ((end = memchr(hdr, ' ', len)) == NULL ||
+		    (end = memchr(hdr, '\t', len)) == NULL)
+			end = hdr + len;
+		l = end - hdr;
+
+		if (len > 8 && !strncmp(hdr, "gwdauth=", 8)) {
+			hdr += 8;
+			len -= 8;
+			l -= 8;
+
+			if (l < MAX_AUTH_COOKIE - 1) {
+				memcpy(params->auth_cookie, hdr, l);
+				params->auth_cookie[l] = '\0';
+			}
+
+			return;
+		}
+
+		/* skip to the next one */
+		hdr += l;
+		len -= l;
+	}
+}
+
 int
 fcgi_parse_params(uint8_t *buf, uint16_t n, struct gotwebd_fcgi_params *params)
 {
@@ -507,6 +561,10 @@ fcgi_parse_params(uint8_t *buf, uint16_t n, struct gotwebd_fcgi_params *params)
 		    strncmp(buf, "HTTPS", 5) == 0)
 			params->https = 1;
 
+		if (name_len == 11 &&
+		    strncmp(buf, "HTTP_COOKIE", 11) == 0)
+			parse_cookie_hdr(params, val, val_len);
+
 		buf += name_len + val_len;
 		n -= name_len - val_len;
 	}
