@@ -43,6 +43,8 @@
 #define GOTWEBD_LOGIN_SOCKET	 "/var/run/gotweb-login.sock"
 #define GOTWEBD_LOGIN_TIMEOUT	 300 /* in seconds */
 
+#define GOTWEBD_CONTROL_SOCKET	 "/var/run/gotwebd.sock"
+
 #define GOTWEBD_MAXDESCRSZ	 1024
 #define GOTWEBD_MAXCLONEURLSZ	 1024
 #define GOTWEBD_CACHESIZE	 1024
@@ -61,16 +63,18 @@
 #define MAX_SERVER_NAME		 255
 #define MAX_AUTH_COOKIE		 255
 #define MAX_IDENTIFIER_SIZE	 32
+#define MAX_BRANCH_NAME		 255
 
 #define GOTWEB_GIT_DIR		 ".git"
 
 #define D_HTTPD_CHROOT		 "/var/www"
+#define D_HTDOCS_PATH		 "/htdocs/gotwebd"
 #define D_UNIX_SOCKET		 "/run/gotweb.sock"
 #define D_FCGI_PORT		 "9000"
 #define D_GOTPATH		 "/got/public"
 #define D_SITENAME		 "Gotweb"
 #define D_SITEOWNER		 "Got Owner"
-#define D_SITELINK		 "Repos"
+#define D_SITELINK		 "Repositories"
 #define D_GOTLOGO		 "got.png"
 #define D_GOTURL		 "https://gameoftrees.org"
 #define D_GOTWEBCSS		 "gotweb.css"
@@ -141,10 +145,14 @@ enum imsg_type {
 	GOTWEBD_IMSG_CFG_SOCK,
 	GOTWEBD_IMSG_CFG_FD,
 	GOTWEBD_IMSG_CFG_ACCESS_RULE,
+	GOTWEBD_IMSG_CFG_MEDIA_TYPE,
 	GOTWEBD_IMSG_CFG_REPO,
+	GOTWEBD_IMSG_CFG_WEBSITE,
 	GOTWEBD_IMSG_CFG_DONE,
 	GOTWEBD_IMSG_CTL_PIPE,
 	GOTWEBD_IMSG_CTL_START,
+	GOTWEBD_IMSG_CTL_STOP,
+	GOTWEBD_IMSG_CTL_INFO,
 	GOTWEBD_IMSG_LOGIN_SECRET,
 	GOTWEBD_IMSG_AUTH_SECRET,
 	GOTWEBD_IMSG_AUTH_CONF,
@@ -165,6 +173,12 @@ struct imsgev {
 
 #define IMSG_DATA_SIZE(imsg)	((imsg)->hdr.len - IMSG_HEADER_SIZE)
 
+/* Structure for GOTWEBD_IMSG_CTL_INFO. */
+struct gotwebd_imsg_info {
+	pid_t pid;
+	int verbosity;
+};
+
 struct env_val {
 	SLIST_ENTRY(env_val)	 entry;
 	char			*val;
@@ -382,6 +396,7 @@ struct gotwebd_repo {
 	TAILQ_ENTRY(gotwebd_repo)	 entry;
 
 	char name[NAME_MAX];
+	char description[GOTWEBD_MAXDESCRSZ];
 
 	enum gotwebd_auth_config	auth_config;
 	struct gotwebd_access_rule_list access_rules;
@@ -390,12 +405,27 @@ struct gotwebd_repo {
 };
 TAILQ_HEAD(gotwebd_repolist, gotwebd_repo);
 
+struct website {
+	STAILQ_ENTRY(website) entry;
+	char repo_name[NAME_MAX];
+	char url_path[MAX_DOCUMENT_URI];
+	char branch_name[MAX_BRANCH_NAME];
+	char path[PATH_MAX];
+
+	enum gotwebd_auth_config	auth_config;
+	struct gotwebd_access_rule_list access_rules;
+};
+
 struct server {
 	TAILQ_ENTRY(server)	 entry;
 
 	char		 name[GOTWEBD_MAXTEXT];
+	char		 htdocs_path[PATH_MAX];
+	char		 gotweb_url_root[MAX_DOCUMENT_URI];
 
 	char		 repos_path[PATH_MAX];
+	char		 repos_url_path[MAX_DOCUMENT_URI];
+	char		 full_repos_url_path[MAX_DOCUMENT_URI * 2 + 2];
 	char		 site_name[GOTWEBD_MAXNAME];
 	char		 site_owner[GOTWEBD_MAXNAME];
 	char		 site_link[GOTWEBD_MAXTEXT];
@@ -403,6 +433,7 @@ struct server {
 	char		 logo_url[GOTWEBD_MAXTEXT];
 	char		 custom_css[PATH_MAX];
 	char		 login_hint_user[MAX_IDENTIFIER_SIZE];
+	char		 login_hint_port[GOTWEBD_MAXPORT];
 
 	size_t		 max_repos_display;
 	size_t		 max_commits_display;
@@ -421,6 +452,7 @@ struct server {
 	struct gotwebd_access_rule_list access_rules;
 
 	struct gotwebd_repolist	 repos;
+	struct got_pathlist_head websites;
 };
 TAILQ_HEAD(serverlist, server);
 
@@ -447,7 +479,6 @@ struct socket {
 	struct socket_conf	 conf;
 
 	int		 fd;
-	struct event	 evt;
 	struct event	 ev;
 	struct event	 pause;
 };
@@ -462,9 +493,15 @@ struct gotwebd {
 	struct socket	*login_sock;
 	struct event	 login_pause_ev;
 
+	struct socket	*control_sock;
+	struct event	 control_pause_ev;
+	struct imsgev	*iev_control;
+
 	enum gotwebd_auth_config auth_config;
 	struct gotwebd_access_rule_list access_rules;
 
+	struct mediatypes mediatypes;
+
 	int		 pack_fds[GOTWEB_PACK_NUM_TEMPFILES];
 	int		 priv_fd[PRIV_FDS__MAX];
 
@@ -489,9 +526,13 @@ struct gotwebd {
 	int		 *worker_load;
 
 	char		 httpd_chroot[PATH_MAX];
+	char		 htdocs_path[PATH_MAX];
+	char		 gotweb_url_root[MAX_DOCUMENT_URI];
+	char		 repos_url_path[MAX_DOCUMENT_URI];
 	uid_t		 www_uid;
 
 	char		 login_hint_user[MAX_IDENTIFIER_SIZE];
+	char		 login_hint_port[GOTWEBD_MAXPORT];
 };
 
 /*
@@ -519,14 +560,14 @@ struct action_keys {
 };
 
 enum querystring_elements {
-	ACTION,
-	COMMIT,
-	RFILE,
-	FOLDER,
-	HEADREF,
-	INDEX_PAGE,
-	PATH,
-	LOGIN,
+	GOTWEBD_QS_ACTION,
+	GOTWEBD_QS_COMMIT,
+	GOTWEBD_QS_RFILE,
+	GOTWEBD_QS_FOLDER,
+	GOTWEBD_QS_HEADREF,
+	GOTWEBD_QS_INDEX_PAGE,
+	GOTWEBD_QS_PATH,
+	GOTWEBD_QS_LOGIN,
 };
 
 extern struct gotwebd	*gotwebd_env;
@@ -570,6 +611,7 @@ void gotwebd_auth(struct gotwebd *, int);
 
 /* gotweb.c */
 struct server *gotweb_get_server(const char *);
+struct website *gotweb_get_website(struct server *, const char *);
 struct gotwebd_repo * gotweb_get_repository(struct server *, const char *);
 int gotweb_reply(struct request *c, int status, const char *ctype,
     struct gotweb_url *);
@@ -582,6 +624,10 @@ int gotweb_render_url(struct request *, struct gotweb_url *);
 int gotweb_render_absolute_url(struct request *, struct gotweb_url *);
 void gotweb_free_repo_commit(struct repo_commit *);
 void gotweb_free_repo_tag(struct repo_tag *);
+void gotweb_log_request(struct request *);
+const struct got_error *gotweb_serve_htdocs(struct request *, const char *);
+const struct got_error *gotweb_route_request(int *, struct website **,
+    char **, struct request *);
 int gotweb_process_request(struct request *);
 void gotweb_free_transport(struct transport *);
 void gotweb(struct gotwebd *, int);
@@ -638,6 +684,7 @@ const struct got_error *got_output_file_blame(struct request *,
 
 /* config.c */
 int config_setserver(struct gotwebd *, struct server *);
+int config_getmediatype(struct gotwebd *, struct imsg *);
 int config_getserver(struct gotwebd *, struct imsg *);
 int config_setsock(struct gotwebd *, struct socket *, uid_t, gid_t);
 int config_getsock(struct gotwebd *, struct imsg *);
@@ -651,4 +698,7 @@ void config_free_access_rules(struct gotwebd_access_rule_list *);
 void config_set_repository(struct imsgev *, struct gotwebd_repo *);
 void config_get_repository(struct gotwebd_repolist *, struct imsg *);
 void config_free_repos(struct gotwebd_repolist *);
+void config_free_websites(struct got_pathlist_head *);
+void config_set_website(struct imsgev *, struct website *);
+void config_get_website(struct got_pathlist_head *, struct imsg *);
 int config_init(struct gotwebd *);
