@@ -38,7 +38,9 @@
 #include "got_error.h"
 #include "got_reference.h"
 #include "got_object.h"
+#include "got_path.h"
 
+#include "media.h"
 #include "gotwebd.h"
 #include "log.h"
 
@@ -53,13 +55,14 @@ struct gotwebd_login_client {
 
 static volatile int client_cnt;
 static int inflight;
+static volatile int stopping;
 
 static char login_token_secret[32];
 
 /*
  * The token format is:
  *
- *    "v1\0"[issued at/64bit][expire/64bit][uid/64bit][host]"\0"
+ *    "v2\0"[random/64bit][issued at/64bit][expire/64bit][uid/64bit][host]"\0"
  *
  * Padded with additional \0 to a length divisible by 4, and then
  * followed by the HMAC-SHA256 of it, all encoded in base64.
@@ -72,7 +75,7 @@ login_check_token(uid_t *euid, char **hostname,
     const char *purpose)
 {
 	time_t	 now;
-	uint64_t issued, expire, uid;
+	uint64_t random, issued, expire, uid;
 	uint8_t *data;
 	int	 len;
 	char	 hmac[32], exp[32];
@@ -109,7 +112,7 @@ login_check_token(uid_t *euid, char **hostname,
 		return -1;
 	}
 
-	if (memcmp(data, "v1", 3) != 0) {
+	if (memcmp(data, "v2", 3) != 0) {
 		log_warnx("unknown %s token format version", purpose);
 		free(data);
 		return -1;
@@ -137,6 +140,9 @@ login_check_token(uid_t *euid, char **hostname,
 		return -1;
 	}
 
+	memcpy(&random, data + used, sizeof(random));
+	used += sizeof(random);
+
 	memcpy(&issued, data + used, sizeof(issued));
 	used += sizeof(issued);
 
@@ -190,7 +196,7 @@ login_gen_token(uint64_t uid, const char *hostname, time_t validity,
 	FILE		*fp;
 	char		*tok;
 	time_t		 now;
-	uint64_t	 issued, expire;
+	uint64_t	 random, issued, expire;
 	size_t		 siz, hlen, pad;
 	unsigned int	 hmaclen;	/* openssl... */
 
@@ -202,6 +208,7 @@ login_gen_token(uint64_t uid, const char *hostname, time_t validity,
 		fatalx("%s token secret length mismatch", purpose);
 
 	now = time(NULL);
+	arc4random_buf(&random, sizeof(random));
 	issued = (uint64_t)now;
 	expire = issued + validity;
 
@@ -212,7 +219,8 @@ login_gen_token(uint64_t uid, const char *hostname, time_t validity,
 	/* include NUL */
 	hlen = strlen(hostname) + 1;
 
-	if (fwrite("v1", 1, 3, fp) != 3 ||
+	if (fwrite("v2", 1, 3, fp) != 3 ||
+	    fwrite(&random, 1, 8, fp) != 8 ||
 	    fwrite(&issued, 1, 8, fp) != 8 ||
 	    fwrite(&expire, 1, 8, fp) != 8 ||
 	    fwrite(&uid, 1, 8, fp) != 8 ||
@@ -411,6 +419,22 @@ login_shutdown(void)
 	exit(0);
 }
 
+static void
+login_stop(void)
+{
+	struct gotwebd *env = gotwebd_env;
+
+	stopping = 1;
+
+	if (env->iev_gotsh) {
+		evtimer_del(&env->login_pause_ev);
+		event_del(&env->iev_gotsh->ev);
+		close(env->iev_gotsh->ibuf.fd);
+		env->iev_gotsh->ibuf.fd = -1;
+		imsgbuf_clear(&env->iev_gotsh->ibuf);
+	}
+}
+
 static void
 login_sighdlr(int sig, short event, void *arg)
 {
@@ -426,8 +450,10 @@ login_sighdlr(int sig, short event, void *arg)
 		break;
 	case SIGCHLD:
 		break;
-	case SIGINT:
 	case SIGTERM:
+		login_stop();
+		break;
+	case SIGINT:
 		login_shutdown();
 		break;
 	default:
@@ -485,7 +511,7 @@ client_read(struct bufferevent *bev, void *d)
 	struct evbuffer *out = EVBUFFER_OUTPUT(bev);
 	char *line, *cmd, *code;
 	size_t linelen;
-	const char *hostname;
+	const char *hostname, *path = NULL;
 
 	if (client->cmd_done) {
 		log_warnx("%s: client sent data even though login command "
@@ -508,12 +534,12 @@ client_read(struct bufferevent *bev, void *d)
 
 	cmd = line;
 	if (strncmp(cmd, "login", 5) == 0) {
+		struct server *srv = NULL;
+
 		cmd += 5;
 		cmd += strspn(cmd, " \t");
 		hostname = cmd;
 		if (hostname[0] == '\0') {
-			struct server *srv;
-
 			/*
 			 * In a multi-server setup we do not want to leak our
 			 * first server's hostname to random people. But if
@@ -522,9 +548,13 @@ client_read(struct bufferevent *bev, void *d)
 			srv = TAILQ_FIRST(&gotwebd_env->servers);
 			if (TAILQ_NEXT(srv, entry) == NULL)
 				hostname = srv->name;
+			else {
+				log_warnx("%s: no hostname provided for "
+				    "weblogin", __func__);
+				client_err(bev, EVBUFFER_READ, client);
+				return;
+			}
 		} else {
-			struct server *srv;
-
 			/* Match hostname against available servers. */
 			TAILQ_FOREACH(srv, &gotwebd_env->servers, entry) {
 				if (strcmp(srv->name, hostname) == 0)
@@ -548,8 +578,12 @@ client_read(struct bufferevent *bev, void *d)
 			return;
 		}
 
-		if (evbuffer_add_printf(out, "ok https://%s/?login=%s\n",
-		    hostname, code) == -1) {
+		if (srv->gotweb_url_root[0] != '\0' &&
+		    !got_path_is_root_dir(srv->gotweb_url_root))
+			path = srv->gotweb_url_root;
+
+		if (evbuffer_add_printf(out, "ok https://%s%s/?login=%s\n",
+		    hostname, path ? path : "", code) == -1) {
 			log_warnx("%s: evbuffer_add_printf failed", __func__);
 			client_err(bev, EVBUFFER_READ, client);
 			free(code);
@@ -601,7 +635,7 @@ login_accept(int fd, short event, void *arg)
 	backoff.tv_sec = 1;
 	backoff.tv_usec = 0;
 
-	if (event_add(&iev->ev, NULL) == -1) {
+	if (!stopping && event_add(&iev->ev, NULL) == -1) {
 		log_warn("event_add");
 		return;
 	}
@@ -622,7 +656,8 @@ login_accept(int fd, short event, void *arg)
 		case EMFILE:
 		case ENFILE:
 			event_del(&iev->ev);
-			evtimer_add(&env->login_pause_ev, &backoff);
+			if (!stopping)
+				evtimer_add(&env->login_pause_ev, &backoff);
 			return;
 		default:
 			log_warn("accept");
@@ -764,6 +799,9 @@ login_dispatch_main(int fd, short event, void *arg)
 		case GOTWEBD_IMSG_CTL_START:
 			login_launch(env);
 			break;
+		case GOTWEBD_IMSG_CTL_STOP:
+			login_stop();
+			break;
 		case GOTWEBD_IMSG_LOGIN_SECRET:
 			if (imsg_get_data(&imsg, login_token_secret,
 			    sizeof(login_token_secret)) == -1)
@@ -791,7 +829,8 @@ accept_paused(int fd, short event, void *arg)
 {
 	struct gotwebd *env = gotwebd_env;
 
-	event_add(&env->iev_gotsh->ev, NULL);
+	if (!stopping)
+		event_add(&env->iev_gotsh->ev, NULL);
 }
 
 void
