@@ -52,7 +52,9 @@
 
 #include "got_reference.h"
 #include "got_object.h"
+#include "got_path.h"
 
+#include "media.h"
 #include "gotwebd.h"
 #include "log.h"
 #include "tmpl.h"
@@ -61,13 +63,13 @@
 #define MAXIMUM(a, b)	(((a) > (b)) ? (a) : (b))
 
 static volatile int client_cnt;
+static volatile int stopping;
 
 static struct timeval	timeout = { TIMEOUT_DEFAULT, 0 };
 
 static void	 sockets_sighdlr(int, short, void *);
 static void	 sockets_shutdown(void);
 static void	 sockets_launch(struct gotwebd *);
-static void	 sockets_accept_paused(int, short, void *);
 
 static void	 sockets_dispatch_main(int, short, void *);
 static int	 sockets_unix_socket_listen(struct gotwebd *, struct socket *, uid_t, gid_t);
@@ -75,9 +77,6 @@ static int	 sockets_create_socket(struct address *);
 static int	 sockets_accept_reserve(int, struct sockaddr *, socklen_t *,
 		    int, volatile int *);
 
-static struct socket *sockets_conf_new_socket(struct gotwebd *,
-		    int, struct address *);
-
 int cgi_inflight = 0;
 
 /* Request hash table needs some spare room to avoid collisions. */
@@ -147,7 +146,8 @@ cleanup_request(struct request *c)
 
 	del_request(c);
 
-	event_add(&c->sock->ev, NULL);
+	if (!stopping)
+		event_add(&c->sock->ev, NULL);
 
 	if (evtimer_initialized(&c->tmo))
 		evtimer_del(&c->tmo);
@@ -157,6 +157,9 @@ cleanup_request(struct request *c)
 		close(c->fd);
 	free(c->buf);
 	free(c);
+
+	if (stopping && client_cnt <= 0)
+		sockets_shutdown();
 }
 
 static void
@@ -235,10 +238,10 @@ sockets(struct gotwebd *env, int fd)
 
 	if (env->prefork <= 0)
 		fatalx("invalid prefork count: %d", env->prefork);
-	env->iev_gotweb = calloc(env->prefork, sizeof(*env->iev_gotweb));
-	if (env->iev_gotweb == NULL)
+	env->iev_auth = calloc(env->prefork, sizeof(*env->iev_auth));
+	if (env->iev_auth == NULL)
 		fatal("calloc");
-	env->gotweb_pending = env->prefork;
+	env->auth_pending = env->prefork;
 
 	signal(SIGPIPE, SIG_IGN);
 
@@ -271,7 +274,7 @@ sockets_parse_sockets(struct gotwebd *env)
 	int sock_id = 1;
 
 	TAILQ_FOREACH(a, &env->addresses, entry) {
-		new_sock = sockets_conf_new_socket(env, sock_id, a);
+		new_sock = sockets_conf_new_socket(sock_id, a);
 		if (new_sock) {
 			sock_id++;
 			TAILQ_INSERT_TAIL(&env->sockets,
@@ -280,8 +283,8 @@ sockets_parse_sockets(struct gotwebd *env)
 	}
 }
 
-static struct socket *
-sockets_conf_new_socket(struct gotwebd *env, int id, struct address *a)
+struct socket *
+sockets_conf_new_socket(int id, struct address *a)
 {
 	struct socket *sock;
 	struct address *acp;
@@ -333,8 +336,8 @@ sockets_launch(struct gotwebd *env)
 
 	if (env->iev_fcgi == NULL)
 		fatalx("fcgi process not connected");
-	if (env->gotweb_pending != 0)
-		fatal("gotweb process not connected");
+	if (env->auth_pending != 0)
+		fatal("auth process not connected");
 
 	TAILQ_FOREACH(sock, &gotwebd_env->sockets, entry) {
 		if (sock->conf.af_type == AF_UNIX) {
@@ -357,7 +360,7 @@ sockets_launch(struct gotwebd *env)
 		if (event_add(&sock->ev, NULL))
 			fatalx("event add sock");
 
-		evtimer_set(&sock->pause, sockets_accept_paused, sock);
+		evtimer_set(&sock->pause, sockets_socket_accept, sock);
 
 		log_info("%s: running socket listener %d", __func__,
 		    sock->conf.id);
@@ -377,7 +380,7 @@ sockets_launch(struct gotwebd *env)
 #endif
 	event_add(&env->iev_fcgi->ev, NULL);
 	for (i = 0; i < env->prefork; i++)
-		event_add(&env->iev_gotweb[i].ev, NULL);
+		event_add(&env->iev_auth[i].ev, NULL);
 }
 
 static void
@@ -399,7 +402,7 @@ abort_request(struct imsg *imsg)
 }
 
 static void
-server_dispatch_gotweb(int fd, short event, void *arg)
+server_dispatch_auth(int fd, short event, void *arg)
 {
 	struct imsgev		*iev = arg;
 	struct imsgbuf		*ibuf;
@@ -448,32 +451,32 @@ server_dispatch_gotweb(int fd, short event, void *arg)
 }
 
 static void
-recv_gotweb_pipe(struct gotwebd *env, struct imsg *imsg)
+recv_auth_pipe(struct gotwebd *env, struct imsg *imsg)
 {
 	struct imsgev *iev;
 	int fd;
 
-	if (env->gotweb_pending <= 0) {
-		log_warn("gotweb pipe already received");
+	if (env->auth_pending <= 0) {
+		log_warn("all auth pipes already received");
 		return;
 	}
 
 	fd = imsg_get_fd(imsg);
 	if (fd == -1)
-		fatalx("invalid gotweb pipe fd");
+		fatalx("invalid auth pipe fd");
 
-	iev = &env->iev_gotweb[env->gotweb_pending - 1];
+	iev = &env->iev_auth[env->auth_pending - 1];
 	if (imsgbuf_init(&iev->ibuf, fd) == -1)
 		fatal("imsgbuf_init");
 	imsgbuf_allow_fdpass(&iev->ibuf);
-	imsgbuf_set_maxsize(&iev->ibuf, sizeof(struct gotwebd_fcgi_record));
+	imsgbuf_set_maxsize(&iev->ibuf, sizeof(struct request));
 
-	iev->handler = server_dispatch_gotweb;
+	iev->handler = server_dispatch_auth;
 	iev->data = iev;
-	event_set(&iev->ev, fd, EV_READ, server_dispatch_gotweb, iev);
+	event_set(&iev->ev, fd, EV_READ, server_dispatch_auth, iev);
 	imsg_event_add(iev);
 
-	env->gotweb_pending--;
+	env->auth_pending--;
 }
 
 static struct imsgev *
@@ -496,15 +499,16 @@ select_worker(struct request *c)
 	    c->request_id, least_busy_worker_idx);
 
 	c->worker_idx = least_busy_worker_idx;
-	return &env->iev_gotweb[least_busy_worker_idx];
+	return &env->iev_auth[least_busy_worker_idx];
 }
 
 static int
 process_request(struct request *c)
 {
 	struct gotwebd *env = gotwebd_env;
-	struct querystring *qs = &c->fcgi_params.qs;
-	struct imsgev *iev_gotweb;
+	struct gotwebd_fcgi_params *params = &c->fcgi_params;
+	struct querystring *qs = &params->qs;
+	struct imsgev *iev_auth;
 	int ret, i;
 	struct request ic;
 
@@ -520,6 +524,16 @@ process_request(struct request *c)
 			return -1;
 		}
 	}
+	if (params->server_name[0] == '\0') {
+		struct server *srv = TAILQ_FIRST(&env->servers);
+
+		if (strlcpy(params->server_name, srv->name,
+		    sizeof(params->server_name)) >=
+		    sizeof(params->server_name)) {
+			log_warnx("server name buffer too small");
+			return -1;
+		}
+	}
 
 	memcpy(&ic, c, sizeof(ic));
 
@@ -536,8 +550,8 @@ process_request(struct request *c)
 		ic.priv_fd[i] = -1;
 	ic.fd = -1;
 
-	iev_gotweb = select_worker(c);
-	ret = imsg_compose_event(iev_gotweb, GOTWEBD_IMSG_REQ_PROCESS,
+	iev_auth = select_worker(c);
+	ret = imsg_compose_event(iev_auth, GOTWEBD_IMSG_REQ_PROCESS,
 	    GOTWEBD_PROC_SOCKETS, -1, c->fd, &ic, sizeof(ic));
 	if (ret == -1) {
 		log_warn("imsg_compose_event");
@@ -546,6 +560,7 @@ process_request(struct request *c)
 	}
 
 	c->fd = -1;
+	event_del(&c->ev);
 	c->client_status = CLIENT_REQUEST;
 	env->worker_load[c->worker_idx]++;
 	return 0;
@@ -615,6 +630,13 @@ recv_parsed_params(struct imsg *imsg)
 		goto fail;
 	}
 
+	if (params.qs.login[0] != '\0' &&
+	    strlcpy(p->qs.login, params.qs.login,
+	    sizeof(p->qs.login)) >= sizeof(p->qs.login)) {
+		log_warnx("login token too long");
+		goto fail;
+	}
+
 	if (params.document_uri[0] != '\0' &&
 	    strlcpy(p->document_uri, params.document_uri,
 	    sizeof(p->document_uri)) >= sizeof(p->document_uri)) {
@@ -629,6 +651,13 @@ recv_parsed_params(struct imsg *imsg)
 		goto fail;
 	}
 
+	if (params.auth_cookie[0] != '\0' &&
+	    strlcpy(p->auth_cookie, params.auth_cookie,
+	    sizeof(p->auth_cookie)) >= sizeof(p->auth_cookie)) {
+		log_warnx("auth cookie too long");
+		goto fail;
+	}
+
 	if (params.https && !p->https)
 		p->https = 1;
 
@@ -710,7 +739,7 @@ recv_fcgi_pipe(struct gotwebd *env, struct imsg *imsg)
 
 	fd = imsg_get_fd(imsg);
 	if (fd == -1)
-		fatalx("invalid gotweb pipe fd");
+		fatalx("invalid fcgi pipe fd");
 
 	iev = calloc(1, sizeof(*iev));
 	if (iev == NULL)
@@ -729,6 +758,32 @@ recv_fcgi_pipe(struct gotwebd *env, struct imsg *imsg)
 	env->iev_fcgi = iev;
 }
 
+static void
+sockets_stop(void)
+{
+	struct socket *sock;
+
+	stopping = 1;
+
+	/*
+	 * Deactivate sockets but to do not free them.
+	 * Client connections maintain a pointer to them via c->sock.
+	 */
+	TAILQ_FOREACH(sock, &gotwebd_env->sockets, entry) {
+		if (sock->fd != -1) {
+			close(sock->fd);
+			sock->fd = -1;
+		}
+		if (event_initialized(&sock->ev))
+			event_del(&sock->ev);
+		if (evtimer_initialized(&sock->ev))
+			evtimer_del(&sock->pause);
+	}
+
+	if (client_cnt == 0)
+		sockets_shutdown();
+}
+
 static void
 sockets_dispatch_main(int fd, short event, void *arg)
 {
@@ -772,11 +827,14 @@ sockets_dispatch_main(int fd, short event, void *arg)
 			if (env->iev_fcgi == NULL)
 				recv_fcgi_pipe(env, &imsg);
 			else
-				recv_gotweb_pipe(env, &imsg);
+				recv_auth_pipe(env, &imsg);
 			break;
 		case GOTWEBD_IMSG_CTL_START:
 			sockets_launch(env);
 			break;
+		case GOTWEBD_IMSG_CTL_STOP:
+			sockets_stop();
+			break;
 		default:
 			fatalx("%s: unknown imsg type %d", __func__,
 			    imsg.hdr.type);
@@ -809,8 +867,10 @@ sockets_sighdlr(int sig, short event, void *arg)
 		break;
 	case SIGCHLD:
 		break;
-	case SIGINT:
 	case SIGTERM:
+		sockets_stop();
+		break;
+	case SIGINT:
 		sockets_shutdown();
 		break;
 	default:
@@ -824,6 +884,8 @@ sockets_shutdown(void)
 {
 	int i;
 
+	requests_purge();
+
 	/* clean servers */
 	while (!TAILQ_EMPTY(&gotwebd_env->servers)) {
 		struct server *srv;
@@ -849,8 +911,6 @@ sockets_shutdown(void)
 		free(sock);
 	}
 
-	requests_purge();
-
 	imsgbuf_clear(&gotwebd_env->iev_parent->ibuf);
 	free(gotwebd_env->iev_parent);
 
@@ -858,8 +918,8 @@ sockets_shutdown(void)
 	free(gotwebd_env->iev_fcgi);
 
 	for (i = 0; i < gotwebd_env->prefork; i++)
-		imsgbuf_clear(&gotwebd_env->iev_gotweb[i].ibuf);
-	free(gotwebd_env->iev_gotweb);
+		imsgbuf_clear(&gotwebd_env->iev_auth[i].ibuf);
+	free(gotwebd_env->iev_auth);
 	free(gotwebd_env->worker_load);
 	free(gotwebd_env);
 
@@ -989,8 +1049,7 @@ sockets_accept_reserve(int sockfd, struct sockaddr *addr, socklen_t *addrlen,
 {
 	int ret;
 
-	if (getdtablecount() + reserve +
-	    ((*counter + 1) * FD_NEEDED) >= getdtablesize()) {
+	if (getdtablecount() + reserve + *counter + 1 >= getdtablesize()) {
 		log_warnx("inflight fds exceeded");
 		errno = EMFILE;
 		return -1;
@@ -1010,14 +1069,6 @@ sockets_accept_reserve(int sockfd, struct sockaddr *addr, socklen_t *addrlen,
 	return ret;
 }
 
-static void
-sockets_accept_paused(int fd, short events, void *arg)
-{
-	struct socket *sock = (struct socket *)arg;
-
-	event_add(&sock->ev, NULL);
-}
-
 static int
 parse_params(struct request *c, uint8_t *record, size_t record_len)
 {
@@ -1163,7 +1214,8 @@ read_fcgi_records(int fd, short events, void *arg)
 		memmove(c->buf, c->buf + record_len, c->buf_len);
 	}
 more:
-	event_add(&c->ev, NULL);
+	if (c->client_status < CLIENT_REQUEST)
+		event_add(&c->ev, NULL);
 	return;
 fail:
 	request_done(c);
@@ -1184,7 +1236,8 @@ sockets_socket_accept(int fd, short event, void *arg)
 	backoff.tv_usec = 0;
 
 	if (event & EV_TIMEOUT) {
-		event_add(&sock->ev, NULL);
+		if (!stopping)
+			event_add(&sock->ev, NULL);
 		return;
 	}
 
@@ -1198,14 +1251,14 @@ sockets_socket_accept(int fd, short event, void *arg)
 		case EINTR:
 		case EWOULDBLOCK:
 		case ECONNABORTED:
-			event_add(&sock->ev, NULL);
+			if (!stopping)
+				event_add(&sock->ev, NULL);
 			return;
 		case EMFILE:
 		case ENFILE:
 			log_warn("accept");
 			event_del(&sock->ev);
-			if (!evtimer_pending(&sock->pause, NULL))
-				evtimer_add(&sock->pause, &backoff);
+			evtimer_add(&sock->pause, &backoff);
 			return;
 		default:
 			log_warn("%s: accept", __func__);
@@ -1217,7 +1270,8 @@ sockets_socket_accept(int fd, short event, void *arg)
 		close(s);
 		if (c != NULL)
 			free(c);
-		event_add(&sock->ev, NULL);
+		if (!stopping)
+			event_add(&sock->ev, NULL);
 		return;
 	}
 
@@ -1226,7 +1280,8 @@ sockets_socket_accept(int fd, short event, void *arg)
 		log_warn("%s: calloc", __func__);
 		close(s);
 		cgi_inflight--;
-		event_add(&sock->ev, NULL);
+		if (!stopping)
+			event_add(&sock->ev, NULL);
 		return;
 	}
 
@@ -1236,7 +1291,8 @@ sockets_socket_accept(int fd, short event, void *arg)
 		close(s);
 		cgi_inflight--;
 		free(c);
-		event_add(&sock->ev, NULL);
+		if (!stopping)
+			event_add(&sock->ev, NULL);
 		return;
 	}
 
