@@ -197,6 +197,10 @@
 #define GOT_ERR_AUTHORIZED_KEY	180
 #define GOT_ERR_CONNECTION_LIMIT 190
 #define GOT_ERR_ON_SERVER_SIDE	191
+#define GOT_ERR_LOGIN_FAILED	192
+#define GOT_ERR_UNKNOWN_COMMAND	193
+#define GOT_ERR_NOT_FOUND	194
+#define GOT_ERR_MEDIA_TYPE	195
 
 struct got_error {
         int code;
